# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Sphinx Configuration."""

project = 'Rosbags'
copyright = '2020-2024, Ternaris'  # noqa: A001
author = 'Ternaris'

autoapi_python_use_implicit_namespaces = True
autodoc_typehints = 'description'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints',
    'sphinx_rtd_theme',
]

html_theme = 'sphinx_rtd_theme'
