# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""CLI tool for rosbag conversion."""

from __future__ import annotations

from .cli import main

if __name__ == '__main__':
    main()
